from .state import State, StateType, state_field, derived, SimulationData

from .reactions import (
    RateType,
    RxnIDType,
    Reaction,
    make_reaction,
    calculate_rates,
    calculate_probabilities,
    ReactionNetworkDef,
    validate_reaction_network,
)

from .model import (
    ODEFunction,
    BaseModelProtocol,
    SupportsDeterministicODE,
    SupportsReactionNetwork,
    SupportsDeterministicSimulation,
    SupportsStochasticSimulation,
)


__all__ = [
    "State",
    "StateType",
    "state_field",
    "derived",
    "SimulationData",
    "RateType",
    "RxnIDType",
    "Reaction",
    "make_reaction",
    "calculate_rates",
    "calculate_probabilities",
    "ReactionNetworkDef",
    "validate_reaction_network",
    "ODEFunction",
    "BaseModelProtocol",
    "SupportsDeterministicODE",
    "SupportsReactionNetwork",
    "SupportsDeterministicSimulation",
    "SupportsStochasticSimulation",
]



from typing import Tuple

import numpy as np

from sparks.utils import dbgprint
from sparks.simulators import DeterministicSimulator
from sparks.data.initiators import dTBPO
from sparks.data.monomers import alpha_methyl_styrene, methyl_methacrylate
from sparks.utils.formulas import C_to_K

from ..model import (
    Model,
    ModelState,
    RateCoefficients,
    SequenceModel,
    SequenceModelState,
)

##################################################
### HIGH TEMPERATURE AMS-MMA (115 C and 140 C) ###
##################################################


def k_AMS_MMA(temp_C: float) -> RateCoefficients:
    """Return the rate coefficients for the AMS MMA model at a given temperature in Celsius."""

    temp_K = C_to_K(temp_C)

    iprops = dTBPO(temp_K)
    mpropsA = alpha_methyl_styrene(temp_K)
    mpropsB = methyl_methacrylate(temp_K)

    kpAA = mpropsA.kp
    kpBB = mpropsB.kp

    rA = 0.16
    rB = 0.41
    kpAB = kpAA / rA
    kpBA = kpBB / rB

    if temp_C == 115:
        KAA = 35
        KBB = 0.22
        qA = 8.2
        qB = 0.0
        kdBA = qA * kpAB
        kdAB = qB * kpBA
        KAB = kdAB / kpAB
        KBA = kdBA / kpBA
    elif temp_C == 140:
        KAA = 54
        KBB = 0.45
        qA = 19.1
        qB = 1.27
        kdBA = qA * kpAB
        kdAB = qB * kpBA
        KAB = kdAB / kpAB
        KBA = kdBA / kpBA
    else:
        raise ValueError(
            f"Unsupported temperature: {temp_C} C. Supported temperatures are 115 C and 140 C."
        )

    return RateCoefficients(
        T_C=temp_C,
        iprops=iprops,
        mpropsA=mpropsA,
        mpropsB=mpropsB,
        rA=rA,
        rB=rB,
        KAA=KAA,
        KAB=KAB,
        KBA=KBA,
        KBB=KBB,
    )


def c0_AMS_MMA(T_C: float, wAMS: float, wDTBPO: float) -> ModelState:
    """Create initial state for AMS-MMA simulation with specified weights."""
    temp_K = C_to_K(T_C)

    iprops = dTBPO(temp_K)
    mpropsA = alpha_methyl_styrene(temp_K)
    mpropsB = methyl_methacrylate(temp_K)

    wMMA = 1 - wAMS

    m_basis = 1000  # Basis in grams
    m_AMS = wAMS * (1 - wDTBPO) * m_basis
    m_MMA = wMMA * (1 - wDTBPO) * m_basis
    m_DTBP = wDTBPO * m_basis

    # Calculate moles
    mol_AMS = m_AMS / mpropsA.MW
    mol_MMA = m_MMA / mpropsB.MW
    mol_DTBP = m_DTBP / iprops.MW

    # Calculate volume in L
    v_AMS = m_AMS / mpropsA.dens_m
    v_MMA = m_MMA / mpropsB.dens_m
    v_total = v_AMS + v_MMA

    V_basis = v_total
    I = mol_DTBP
    A = mol_AMS
    B = mol_MMA

    return ModelState(V=V_basis, I=I, A=A, B=B)


# ---- Generate data for validation ----

solver_kwargs = dict(method="BDF", rtol=1e-8, atol=1e-8)


def get_sim_data() -> Tuple[DeterministicSimulator, DeterministicSimulator]:

    T_C = 115
    ds_115 = DeterministicSimulator(Model(), k=k_AMS_MMA(T_C))
    data_115 = ds_115.simulate(
        t_eval=np.linspace(0, 1450, 1000),
        init_state=c0_AMS_MMA(T_C, wAMS=0.45, wDTBPO=0.02),
        solver_kwargs=solver_kwargs,
    )

    T_C = 140
    ds_140 = DeterministicSimulator(Model(), k=k_AMS_MMA(T_C))
    data_140 = ds_140.simulate(
        t_eval=np.linspace(0, 1080, 1000),
        init_state=c0_AMS_MMA(T_C, wAMS=0.45, wDTBPO=0.02),
        solver_kwargs=solver_kwargs,
    )

    return ds_115, ds_140


def get_sim_sequence_data(
    ds_115: DeterministicSimulator, ds_140: DeterministicSimulator
) -> Tuple[DeterministicSimulator, DeterministicSimulator]:

    T_C = 115
    ds_115s = DeterministicSimulator(SequenceModel(), k=k_AMS_MMA(T_C))
    data_115 = ds_115s.simulate(
        t_eval=np.linspace(0, 1450, 1000),
        init_state=SequenceModelState(),
        solver_kwargs=solver_kwargs,
        s=ds_115.ode_sol,
    )

    T_C = 140
    ds_140s = DeterministicSimulator(SequenceModel(), k=k_AMS_MMA(T_C))
    data_140 = ds_140s.simulate(
        t_eval=np.linspace(0, 1080, 1000),
        init_state=SequenceModelState(),
        solver_kwargs=solver_kwargs,
        s=ds_140.ode_sol,
    )

    return ds_115s, ds_140s



from .formulas import (
    arrhenius_eqn,
    get_arrhenius_const,
    cal_to_J,
    kcal_to_J,
    C_to_K,
    K_to_C,
)

from .print import okprint, errprint, warnprint, dbgprint

__all__ = [
    "arrhenius_eqn",
    "get_arrhenius_const",
    "cal_to_J",
    "kcal_to_J",
    "C_to_K",
    "K_to_C",
    "okprint",
    "errprint",
    "warnprint",
    "dbgprint",
]



from __future__ import annotations
from typing import Type

import numpy as np

from sparks.core import NDArrayType
from sparks.models import (
    RateType,
    StateType,
    ReactionNetworkDef,
    calculate_rates,
    ODEFunction,
)


def optimized_reaction_odes(
    state_class: Type[StateType], reactions: ReactionNetworkDef
) -> ODEFunction:
    """
    Returns an optimized ODE function that pre-computes field mappings.
    """

    # Pre-compute mappings once (closure captures these)
    valid_fields = state_class._data_fields()
    field_map = {f.name: i for i, f in enumerate(valid_fields)}
    print(f"Field map: {field_map}")

    print(f"All possible fields: {list(field_map.keys())}")

    # Check which fields are actually used in reactions
    used_fields = set()
    for rxn_id, reaction in reactions.items():
        used_fields.update(reaction.reactants.keys())
        used_fields.update(reaction.products.keys())
    print(f"Fields used in reactions: {used_fields}")

    # Pre-process reactions for faster access
    reaction_data = []
    for rxn_id, reaction in reactions.items():
        reactant_indices = [
            (field_map[species], stoich)
            for species, stoich in reaction.reactants.items()
            if species in field_map
        ]

        product_indices = [
            (field_map[species], stoich)
            for species, stoich in reaction.products.items()
            if species in field_map
        ]

        reaction_data.append((rxn_id, reaction, reactant_indices, product_indices))

    def optimized_ode_function(
        k: RateType,
        t: NDArrayType,
        state: StateType,
    ) -> StateType:
        """The actual optimized ODE function."""
        # Initialize derivative array
        dydt_array = np.zeros(len(field_map))

        # Calculate rates
        rates = calculate_rates(state, k, reactions)

        # Apply stoichiometry using pre-computed indices
        for rxn_id, reaction, reactant_indices, product_indices in reaction_data:
            rate_value = rates[rxn_id]

            # Consumption
            for species_idx, stoich in reactant_indices:
                dydt_array[species_idx] -= stoich * rate_value

            # Formation
            for species_idx, stoich in product_indices:
                dydt_array[species_idx] += stoich * rate_value

        # Convert back to state object
        return state_class.from_array(dydt_array, t=t)

    return optimized_ode_function



from typing import Any, Dict, List, Tuple

import numpy as np
from runkmc.results import SimulationResult

from sparks.models import SimulationData
from sparks.simulators.stochastic import KMCConfig
from .model import ModelState, RateCoefficients, ChainModelState, SequenceModelState


def get_FRP2_kmc_inputs(
    k: RateCoefficients, s0: ModelState, config: KMCConfig
) -> Dict[str, Any]:

    return {
        "num_units": config.num_units,
        "termination_time": config.termination_time,
        "analysis_time": config.analysis_time,
        "I_c0": float(s0.I),
        "R_c0": float(s0.R),
        "A_c0": float(s0.A),
        "B_c0": float(s0.B),
        "I_FW": 0.0,
        "R_FW": 1.0,
        "A_FW": 100.0,
        "B_FW": 200.0,
        "kd": k.kd,
        "kpAA": k.kpAA,
        "kpAB": k.kpAB,
        "kpBA": k.kpBA,
        "kpBB": k.kpBB,
        "kdAA": k.kdAA,
        "kdAB": k.kdAB,
        "kdBA": k.kdBA,
        "kdBB": k.kdBB,
        "ktcAA": k.ktcAA,
        "ktcAB": k.ktcAB,
        "ktcBB": k.ktcBB,
        "ktdAA": k.ktdAA,
        "ktdAB": k.ktdAB,
        "ktdBB": k.ktdBB,
    }


def get_CRP3_kmc_inputs(
    k: RateCoefficients, s0: ModelState, config: KMCConfig
) -> Dict[str, Any]:

    return {
        "num_units": config.num_units,
        "termination_time": config.termination_time,
        "analysis_time": config.analysis_time,
        "R_c0": float(s0.R),
        "A_c0": float(s0.A),
        "B_c0": float(s0.B),
        "R_FW": 1.0,
        "A_FW": 100.0,
        "B_FW": 200.0,
        "kpAA": k.kpAA,
        "kpAB": k.kpAB,
        "kpBA": k.kpBA,
        "kpBB": k.kpBB,
        "kdAA": k.kdAA,
        "kdAB": k.kdAB,
        "kdBA": k.kdBA,
        "kdBB": k.kdBB,
    }


def get_kmc_inputs(
    k: RateCoefficients, s0: ModelState, config: KMCConfig, model_name: str
) -> Dict[str, Any]:

    if model_name == "CRP3":
        return get_CRP3_kmc_inputs(k, s0, config)
    elif model_name == "FRP2":
        return get_FRP2_kmc_inputs(k, s0, config)
    else:
        raise ValueError(f"Unknown model name: {model_name}")


def build_chain_model_state(
    results: SimulationResult,
) -> SimulationData[ChainModelState]:

    data = results.results
    t = data.kmc_time
    zeros = np.zeros_like(t)

    cms = ChainModelState(
        t=t,
        lam_0=zeros,
        lam_1=zeros,
        lam_2=zeros,
        mu_0=np.ones_like(t),
        mu_1=data.nAvgCL,
        mu_2=data.wAvgCL * data.nAvgCL,
    )

    return SimulationData[ChainModelState].from_trajectory(cms)


def build_sequence_model_state(
    results: SimulationResult,
) -> SimulationData[SequenceModelState]:

    data = results.results
    t = data.kmc_time
    zeros = np.zeros_like(t)

    sms = SequenceModelState(
        t=t,
        aSA0=zeros,
        aSB0=zeros,
        aSA1=zeros,
        aSB1=zeros,
        aSA2=zeros,
        aSB2=zeros,
        iSA0=np.ones_like(t),
        iSB0=np.ones_like(t),
        iSA1=data.nAvgSL["A"],
        iSB1=data.nAvgSL["B"],
        iSA2=data.wAvgSL["A"] * data.nAvgSL["A"],
        iSB2=data.wAvgSL["B"] * data.nAvgSL["B"],
    )

    return SimulationData[SequenceModelState].from_trajectory(sms)


def build_pos_sequence_model_state(
    results: SimulationResult,
) -> List[SimulationData[SequenceModelState]]:

    from runkmc.results import SequenceData

    seq_data = results.sequence_data
    if seq_data is None:
        print("No sequence data found in results.")
        return []

    pos_sms = []
    buckets = sorted(np.unique(seq_data.bucket))

    buckets = seq_data.get_buckets()

    for pos in buckets:

        _seq_data = seq_data.get_by_bucket(pos)

        assert isinstance(_seq_data, SequenceData)

        t = _seq_data.kmc_time
        zeros = np.zeros_like(t)
        sms = SequenceModelState(
            t=t,
            aSA0=zeros,
            aSB0=zeros,
            aSA1=zeros,
            aSB1=zeros,
            aSA2=zeros,
            aSB2=zeros,
            iSA0=np.ones_like(t),
            iSB0=np.ones_like(t),
            iSA1=_seq_data.nAvgSL["A"],
            iSB1=_seq_data.nAvgSL["B"],
            iSA2=_seq_data.wAvgSL["A"] * _seq_data.nAvgSL["A"],
            iSB2=_seq_data.wAvgSL["B"] * _seq_data.nAvgSL["B"],
        )
        sms_data = SimulationData[SequenceModelState].from_trajectory(sms)
        pos_sms.append(sms_data)

    return pos_sms


def build_FRP2_state(results: SimulationResult) -> SimulationData[ModelState]:

    data = results.results

    NAV = data.NAV
    state = ModelState(
        t=data.kmc_time,
        I=data.unit_counts["I"] / NAV,
        R=data.unit_counts["R"] / NAV,
        A=data.unit_counts["A"] / NAV,
        B=data.unit_counts["B"] / NAV,
        RA=data.polymer_counts["P[R.A]"] / NAV,
        RB=data.polymer_counts["P[R.B]"] / NAV,
        PAA=data.polymer_counts["P[A.A]"] / NAV,
        PAB=data.polymer_counts["P[A.B]"] / NAV,
        PBA=data.polymer_counts["P[B.A]"] / NAV,
        PBB=data.polymer_counts["P[B.B]"] / NAV,
        PD=data.polymer_counts["D"] / NAV,
    )

    return SimulationData[ModelState].from_trajectory(state)


def build_CRP3_state(results: SimulationResult) -> SimulationData[ModelState]:

    data = results.results

    NAV = data.NAV
    state = ModelState(
        t=data.kmc_time,
        I=np.zeros_like(NAV),
        R=data.unit_counts["R"] / NAV,
        A=data.unit_counts["A"] / NAV,
        B=data.unit_counts["B"] / NAV,
        RA=data.polymer_counts["P[R.A]"] / NAV,
        RB=data.polymer_counts["P[R.B]"] / NAV,
        PAA=data.polymer_counts["P[-.A.A]"] / NAV,
        PAB=data.polymer_counts["P[-.A.B]"] / NAV,
        PBA=data.polymer_counts["P[-.B.A]"] / NAV,
        PBB=data.polymer_counts["P[-.B.B]"] / NAV,
        PD=data.polymer_counts["D"] / NAV,
    )

    return SimulationData[ModelState].from_trajectory(state)


def parse_kmc_outputs(
    result: SimulationResult, model_name: str
) -> Tuple[SimulationData[ModelState], Any]:

    model_data = None
    if model_name == "FRP2":
        model_data = build_FRP2_state(result)
    elif model_name == "CRP3":
        model_data = build_CRP3_state(result)
    else:
        raise ValueError(f"Unknown model name: {model_name}")

    cms = build_chain_model_state(result)
    sms = build_sequence_model_state(result)

    pos_sms = None
    if result.sequence_data is not None:
        pos_sms = build_pos_sequence_model_state(result)

    mdata = {"cms": cms, "sms": sms, "pos_sms": pos_sms}

    return model_data, mdata



from enum import Enum, auto

from sparks.models import ReactionNetworkDef, make_reaction


class RxnID(Enum):
    """All possible reactions in the copolymerization system"""

    INIT_DECOMP = auto()
    THERM_INIT_A = auto()
    THERM_INIT_B = auto()

    PROP_R_A = auto()
    PROP_R_B = auto()
    PROP_RA_A = auto()
    PROP_RA_B = auto()
    PROP_RB_A = auto()
    PROP_RB_B = auto()

    PROP_PAA_A = auto()
    PROP_PAA_B = auto()
    PROP_PAB_A = auto()
    PROP_PAB_B = auto()
    PROP_PBA_A = auto()
    PROP_PBA_B = auto()
    PROP_PBB_A = auto()
    PROP_PBB_B = auto()

    DEPROP_PAAA = auto()
    DEPROP_PAAB = auto()
    DEPROP_PABA = auto()
    DEPROP_PABB = auto()
    DEPROP_PBAA = auto()
    DEPROP_PBAB = auto()
    DEPROP_PBBA = auto()
    DEPROP_PBBB = auto()

    TERMC_PAA_PAA = auto()
    TERMC_PAA_PBA = auto()
    TERMC_PBA_PBA = auto()

    TERMC_PAA_PAB = auto()
    TERMC_PAA_PBB = auto()
    TERMC_PBA_PAB = auto()
    TERMC_PBA_PBB = auto()

    TERMC_PAB_PAB = auto()
    TERMC_PAB_PBB = auto()
    TERMC_PBB_PBB = auto()

    TERMD_PAA_PAA = auto()
    TERMD_PAA_PBA = auto()
    TERMD_PBA_PBA = auto()

    TERMD_PAA_PAB = auto()
    TERMD_PAA_PBB = auto()
    TERMD_PBA_PAB = auto()
    TERMD_PBA_PBB = auto()

    TERMD_PAB_PAB = auto()
    TERMD_PAB_PBB = auto()
    TERMD_PBB_PBB = auto()

    CTR_PAA_A = auto()
    CTR_PAA_B = auto()
    CTR_PAB_A = auto()
    CTR_PAB_B = auto()
    CTR_PBA_A = auto()
    CTR_PBA_B = auto()
    CTR_PBB_A = auto()
    CTR_PBB_B = auto()


def initiator_decomposition_rxn() -> ReactionNetworkDef:

    from .model import RateCoefficients, ModelState

    state = ModelState()
    Reaction = make_reaction(RateCoefficients, ModelState)

    return {
        # Initiation
        RxnID.INIT_DECOMP: Reaction(
            reactants={"I": 1},
            products={"R": 2 * 0.5},
            rate=lambda k, s: k.kd * s.I,
        ),
    }


def thermal_initiation_rxn() -> ReactionNetworkDef:

    from .model import RateCoefficients, ModelState

    state = ModelState()
    Reaction = make_reaction(RateCoefficients, ModelState)

    return {
        RxnID.THERM_INIT_A: Reaction(
            reactants={"A": 3},
            products={"RA": 2},
            rate=lambda k, s: k.kthA * (s.A**3),
        ),
        RxnID.THERM_INIT_B: Reaction(
            reactants={"B": 3},
            products={"RB": 2},
            rate=lambda k, s: k.kthB * (s.B**3),
        ),
    }


def propagation_rxn() -> ReactionNetworkDef:

    from .model import RateCoefficients, ModelState

    state = ModelState()
    Reaction = make_reaction(RateCoefficients, ModelState)

    return {
        RxnID.PROP_R_A: Reaction(
            reactants={"R": 1, "A": 1},
            products={"RA": 1},
            rate=lambda k, s: k.kpAA * s.R * s.A,
        ),
        RxnID.PROP_R_B: Reaction(
            reactants={"R": 1, "B": 1},
            products={"RB": 1},
            rate=lambda k, s: k.kpBB * s.R * s.B,
        ),
        RxnID.PROP_RA_A: Reaction(
            reactants={"RA": 1, "A": 1},
            products={"PAA": 1},
            rate=lambda k, s: k.kpAA * s.RA * s.A,
        ),
        RxnID.PROP_RA_B: Reaction(
            reactants={"RA": 1, "B": 1},
            products={"PAB": 1},
            rate=lambda k, s: k.kpAB * s.RA * s.B,
        ),
        RxnID.PROP_RB_A: Reaction(
            reactants={"RB": 1, "A": 1},
            products={"PBA": 1},
            rate=lambda k, s: k.kpBA * s.RB * s.A,
        ),
        RxnID.PROP_RB_B: Reaction(
            reactants={"RB": 1, "B": 1},
            products={"PBB": 1},
            rate=lambda k, s: k.kpBB * s.RB * s.B,
        ),
        RxnID.PROP_PAA_A: Reaction(
            reactants={"PAA": 1, "A": 1},
            products={"PAA": 1},
            rate=lambda k, s: k.kpAA * s.PAA * s.A,
        ),
        RxnID.PROP_PAA_B: Reaction(
            reactants={"PAA": 1, "B": 1},
            products={"PAB": 1},
            rate=lambda k, s: k.kpAB * s.PAA * s.B,
        ),
        RxnID.PROP_PAB_A: Reaction(
            reactants={"PAB": 1, "A": 1},
            products={"PBA": 1},
            rate=lambda k, s: k.kpBA * s.PAB * s.A,
        ),
        RxnID.PROP_PAB_B: Reaction(
            reactants={"PAB": 1, "B": 1},
            products={"PBB": 1},
            rate=lambda k, s: k.kpBB * s.PAB * s.B,
        ),
        RxnID.PROP_PBA_A: Reaction(
            reactants={"PBA": 1, "A": 1},
            products={"PAA": 1},
            rate=lambda k, s: k.kpAA * s.PBA * s.A,
        ),
        RxnID.PROP_PBA_B: Reaction(
            reactants={"PBA": 1, "B": 1},
            products={"PAB": 1},
            rate=lambda k, s: k.kpAB * s.PBA * s.B,
        ),
        RxnID.PROP_PBB_A: Reaction(
            reactants={"PBB": 1, "A": 1},
            products={"PBA": 1},
            rate=lambda k, s: k.kpBA * s.PBB * s.A,
        ),
        RxnID.PROP_PBB_B: Reaction(
            reactants={"PBB": 1, "B": 1},
            products={"PBB": 1},
            rate=lambda k, s: k.kpBB * s.PBB * s.B,
        ),
    }


def depropagation_rxn() -> ReactionNetworkDef:

    from .model import RateCoefficients, ModelState

    Reaction = make_reaction(RateCoefficients, ModelState)

    return {
        RxnID.DEPROP_PAAA: Reaction(
            reactants={"PAA": 1},
            products={"PAA": 1, "A": 1},
            rate=lambda k, s: k.kdAA * s.pAA * s.PAA,
        ),
        RxnID.DEPROP_PAAB: Reaction(
            reactants={"PAB": 1},
            products={"PAA": 1, "B": 1},
            rate=lambda k, s: k.kdAB * s.pAA * s.PAB,
        ),
        RxnID.DEPROP_PABA: Reaction(
            reactants={"PBA": 1},
            products={"PAB": 1, "A": 1},
            rate=lambda k, s: k.kdBA * s.pAB * s.PBA,
        ),
        RxnID.DEPROP_PABB: Reaction(
            reactants={"PBB": 1},
            products={"PAB": 1, "B": 1},
            rate=lambda k, s: k.kdBB * s.pAB * s.PBB,
        ),
        RxnID.DEPROP_PBAA: Reaction(
            reactants={"PAA": 1},
            products={"PBA": 1, "A": 1},
            rate=lambda k, s: k.kdAA * s.pBA * s.PAA,
        ),
        RxnID.DEPROP_PBAB: Reaction(
            reactants={"PAB": 1},
            products={"PBA": 1, "B": 1},
            rate=lambda k, s: k.kdAB * s.pBA * s.PAB,
        ),
        RxnID.DEPROP_PBBA: Reaction(
            reactants={"PBA": 1},
            products={"PBB": 1, "A": 1},
            rate=lambda k, s: k.kdBA * s.pBB * s.PBA,
        ),
        RxnID.DEPROP_PBBB: Reaction(
            reactants={"PBB": 1},
            products={"PBB": 1, "B": 1},
            rate=lambda k, s: k.kdBB * s.pBB * s.PBB,
        ),
    }


def termination_rxn() -> ReactionNetworkDef:

    from .model import RateCoefficients, ModelState

    state = ModelState()
    Reaction = make_reaction(RateCoefficients, ModelState)

    return {
        # Termination (combination)
        RxnID.TERMC_PAA_PAA: Reaction(
            reactants={"PAA": 2},
            products={"PD": 1},
            rate=lambda k, s: 0.5 * k.ktcAA * s.PAA * s.PAA,
        ),
        RxnID.TERMC_PAA_PBA: Reaction(
            reactants={"PAA": 1, "PBA": 1},
            products={"PD": 1},
            rate=lambda k, s: 0.5 * k.ktcAA * s.PAA * s.PBA,
        ),
        RxnID.TERMC_PBA_PBA: Reaction(
            reactants={"PBA": 2},
            products={"PD": 1},
            rate=lambda k, s: 0.5 * k.ktcAA * s.PBA * s.PBA,
        ),
        RxnID.TERMC_PAA_PAB: Reaction(
            reactants={"PAA": 1, "PAB": 1},
            products={"PD": 1},
            rate=lambda k, s: 0.5 * k.ktcAB * s.PAA * s.PAB,
        ),
        RxnID.TERMC_PAA_PBB: Reaction(
            reactants={"PAA": 1, "PBB": 1},
            products={"PD": 1},
            rate=lambda k, s: 0.5 * k.ktcAB * s.PAA * s.PBB,
        ),
        RxnID.TERMC_PBA_PAB: Reaction(
            reactants={"PBA": 1, "PAB": 1},
            products={"PD": 1},
            rate=lambda k, s: 0.5 * k.ktcAB * s.PBA * s.PAB,
        ),
        RxnID.TERMC_PBA_PBB: Reaction(
            reactants={"PBA": 1, "PBB": 1},
            products={"PD": 1},
            rate=lambda k, s: 0.5 * k.ktcAB * s.PBA * s.PBB,
        ),
        RxnID.TERMC_PAB_PAB: Reaction(
            reactants={"PAB": 2},
            products={"PD": 1},
            rate=lambda k, s: 0.5 * k.ktcBB * s.PAB * s.PAB,
        ),
        RxnID.TERMC_PAB_PBB: Reaction(
            reactants={"PAB": 1, "PBB": 1},
            products={"PD": 1},
            rate=lambda k, s: 0.5 * k.ktcBB * s.PAB * s.PBB,
        ),
        RxnID.TERMC_PBB_PBB: Reaction(
            reactants={"PBB": 2},
            products={"PD": 1},
            rate=lambda k, s: 0.5 * k.ktcBB * s.PBB * s.PBB,
        ),
        # Termination (disproportionation)
        RxnID.TERMD_PAA_PAA: Reaction(
            reactants={"PAA": 2},
            products={"PD": 2},
            rate=lambda k, s: k.ktdAA * s.PAA * s.PAA,
        ),
        RxnID.TERMD_PAA_PBA: Reaction(
            reactants={"PAA": 1, "PBA": 1},
            products={"PD": 2},
            rate=lambda k, s: k.ktdAA * s.PAA * s.PBA,
        ),
        RxnID.TERMD_PBA_PBA: Reaction(
            reactants={"PBA": 2},
            products={"PD": 2},
            rate=lambda k, s: k.ktdAA * s.PBA * s.PBA,
        ),
        RxnID.TERMD_PAA_PAB: Reaction(
            reactants={"PAA": 1, "PAB": 1},
            products={"PD": 2},
            rate=lambda k, s: k.ktdAB * s.PAA * s.PAB,
        ),
        RxnID.TERMD_PAA_PBB: Reaction(
            reactants={"PAA": 1, "PBB": 1},
            products={"PD": 2},
            rate=lambda k, s: k.ktdAB * s.PAA * s.PBB,
        ),
        RxnID.TERMD_PBA_PAB: Reaction(
            reactants={"PBA": 1, "PAB": 1},
            products={"PD": 2},
            rate=lambda k, s: k.ktdAB * s.PBA * s.PAB,
        ),
        RxnID.TERMD_PBA_PBB: Reaction(
            reactants={"PBA": 1, "PBB": 1},
            products={"PD": 2},
            rate=lambda k, s: k.ktdAB * s.PBA * s.PBB,
        ),
        RxnID.TERMD_PAB_PAB: Reaction(
            reactants={"PAB": 2},
            products={"PD": 2},
            rate=lambda k, s: k.ktdBB * s.PAB * s.PAB,
        ),
        RxnID.TERMD_PAB_PBB: Reaction(
            reactants={"PAB": 1, "PBB": 1},
            products={"PD": 2},
            rate=lambda k, s: k.ktdBB * s.PAB * s.PBB,
        ),
        RxnID.TERMD_PBB_PBB: Reaction(
            reactants={"PBB": 2},
            products={"PD": 2},
            rate=lambda k, s: k.ktdBB * s.PBB * s.PBB,
        ),
    }


def chain_transfer_rxn() -> ReactionNetworkDef:

    from .model import RateCoefficients, ModelState

    state = ModelState()
    Reaction = make_reaction(RateCoefficients, ModelState)

    return {
        # Chain Transfer to Monomer
        RxnID.CTR_PAA_A: Reaction(
            reactants={"PAA": 1, "A": 1},
            products={"PD": 1, "RA": 1},
            rate=lambda k, s: k.ktrAA * s.PAA * s.A,
        ),
        RxnID.CTR_PAA_B: Reaction(
            reactants={"PAA": 1, "B": 1},
            products={"PD": 1, "RB": 1},
            rate=lambda k, s: k.ktrAB * s.PAA * s.B,
        ),
        RxnID.CTR_PAB_A: Reaction(
            reactants={"PAB": 1, "A": 1},
            products={"PD": 1, "RA": 1},
            rate=lambda k, s: k.ktrBA * s.PAB * s.A,
        ),
        RxnID.CTR_PAB_B: Reaction(
            reactants={"PAB": 1, "B": 1},
            products={"PD": 1, "RB": 1},
            rate=lambda k, s: k.ktrBB * s.PAB * s.B,
        ),
        RxnID.CTR_PBA_A: Reaction(
            reactants={"PBA": 1, "A": 1},
            products={"PD": 1, "RA": 1},
            rate=lambda k, s: k.ktrAA * s.PBA * s.A,
        ),
        RxnID.CTR_PBA_B: Reaction(
            reactants={"PBA": 1, "B": 1},
            products={"PD": 1, "RB": 1},
            rate=lambda k, s: k.ktrAB * s.PBA * s.B,
        ),
        RxnID.CTR_PBB_A: Reaction(
            reactants={"PBB": 1, "A": 1},
            products={"PD": 1, "RA": 1},
            rate=lambda k, s: k.ktrBA * s.PBB * s.A,
        ),
        RxnID.CTR_PBB_B: Reaction(
            reactants={"PBB": 1, "B": 1},
            products={"PD": 1, "RB": 1},
            rate=lambda k, s: k.ktrBB * s.PBB * s.B,
        ),
    }


def reactions() -> ReactionNetworkDef:
    """Combine all reaction definitions into a single network."""
    return {
        **initiator_decomposition_rxn(),
        **thermal_initiation_rxn(),
        **propagation_rxn(),
        **depropagation_rxn(),
        **termination_rxn(),
        **chain_transfer_rxn(),
    }



def _colorprint(print_str: str, color_code: str):
    """Prints a message in the specified ANSI color."""
    print(f"\x1b[{color_code}m{print_str}\x1b[0m")


def okprint(print_str: str):
    """Prints a message in green."""
    _colorprint(print_str, "92")


def errprint(print_str: str):
    """Prints error message in red."""
    _colorprint(print_str, "91")


def warnprint(print_str: str):
    """Prints warning message in pale orange."""
    _colorprint(print_str, "38;5;215")


def dbgprint(print_str: str):
    """Prints debug message in light blue."""
    _colorprint(print_str, "96")



from dataclasses import dataclass

import numpy as np

from sparks.core import NDArrayType
from sparks.data import MonomerType

from .model import RateCoefficients, ModelState, SequenceModelState


def get_kt_from_Cheong_2003(s: ModelState, k: RateCoefficients):

    T_C = k.T_C
    if T_C == 115:
        ka = 0.23e14  # L/mol/min
    elif T_C == 140:
        ka = 0.8e14  # L/mol/min
    else:
        raise ValueError(f"Unsupported temperature: {T_C} C")

    ktBB = k.mpropsB.kt
    ktAA = (np.sqrt(ka) - s.beta * np.sqrt(ktBB)) ** 2
    ktAB = np.sqrt(ktAA * ktBB)

    return ktAA, ktAB, ktBB


@dataclass
class Context:

    init_state: ModelState | None = None


def kinetic_odes(
    dydt: ModelState,
    k: RateCoefficients,
    t: NDArrayType,
    s: ModelState,
    context: Context = Context(),
) -> ModelState:
    """Defines the ODEs for the full FRP2 model."""

    dydt.t = t

    T_C = k.T_C
    T_K = T_C + 273.15  # Convert Celsius to Kelvin

    # Extract properties
    iprops = k.iprops
    mpropsA = k.mpropsA
    mpropsB = k.mpropsB

    # Save initial state
    if not context.init_state and t >= 0:
        context.init_state = s

    # A0, B0 are initial moles for volume balance calculations
    A0 = context.init_state.A
    B0 = context.init_state.B
    M0 = A0 + B0

    ######################
    ### Rate Constants ###
    ######################

    kpAA, kpAB, kpBA, kpBB = k.kp()
    kpRAA, kpRAB, kpRBA, kpRBB = k.kp(f=1e1)
    kdAA, kdAB, kdBA, kdBB = k.kd()

    # Initiation
    kd = iprops.kd[MonomerType.General]
    f = iprops.f[MonomerType.General]

    # Thermal Initiation
    kthA = mpropsA.kth
    kthB = mpropsB.kth

    # Chain Transfer to Monomer
    kf_mAA, kf_mAB, kf_mBA, kf_mBB = k.kfm()
    kfm = kfm_pseudo = (
        kf_mAA * s.pA * s.fA
        + kf_mAB * s.pA * s.fB
        + kf_mBA * s.pB * s.fA
        + kf_mBB * s.pB * s.fB
    )

    RpA = (
        s.A_conc
        * (
            kpRAA * (s.R_conc + s.RA_conc)
            + kpRBA * s.RB_conc
            + kpAA * s.PA_conc
            + kpBA * s.PB_conc
        )
        - s.PA_conc * (kdAA * s.pAA + kdBA * s.pBA)
        + s.A_conc * (kf_mAA * s.PA_conc + kf_mBA * s.PB_conc)
        + 2 * kthA * s.A_conc**3
    )
    RpB = (
        s.B_conc
        * (
            kpRBB * (s.R_conc + s.RB_conc)
            + kpRAB * s.RA_conc
            + kpBB * s.PB_conc
            + kpAB * s.PA_conc
        )
        - s.PB_conc * (kdBB * s.pBB + kdAB * s.pAB)
        + s.B_conc * (kf_mAB * s.PA_conc + kf_mBB * s.PB_conc)
        + 2 * kthB * s.B_conc**3
    )

    ###########################################
    ### Calculate the pseudo rate constants ###
    ###########################################

    # Pseudo rate constant for propagation
    kp = (
        # i=A, j=A
        (kpAA * s.pA * s.fA - kdAA * s.pAA * s.pA / s.M_conc)
        # i=A, j=B
        + (kpAB * s.pA * s.fB - kdAB * s.pAB * s.pB / s.M_conc)
        # i=B, j=A
        + (kpBA * s.pB * s.fA - kdBA * s.pBA * s.pA / s.M_conc)
        # i=B, j=B
        + (kpBB * s.pB * s.fB - kdBB * s.pBB * s.pB / s.M_conc)
    )

    ktAA, ktAB, ktBB = get_kt_from_Cheong_2003(s, k)
    kt = ktAA * s.pA * s.pA + ktBB * s.pB * s.pB + 2 * ktAB * s.pA * s.pB
    ktd_ratio = mpropsA.ktd_ratio * s.pA + mpropsB.ktd_ratio * s.pB

    ########################################
    ### Calculate the overall properties ###
    ########################################

    # Calculate the mass/volume of monomer
    m_A = s.A * mpropsA.MW
    m_B = s.B * mpropsB.MW
    vol_A = m_A / mpropsA.dens_m
    vol_B = m_B / mpropsB.dens_m
    vol_m = vol_A + vol_B

    # Calculate the mass/volume of polymer
    m_PA = s.nPA * mpropsA.MW
    m_PB = s.nPB * mpropsB.MW
    m_P = m_PA + m_PB
    vol_PA = m_PA / mpropsA.dens_p
    vol_PB = m_PB / mpropsB.dens_p
    vol_P = vol_PA + vol_PB  # Volume of polymer in L

    ktc = kt / (1 + ktd_ratio)
    ktd = kt * ktd_ratio / (1 + ktd_ratio)

    # Calculate volume change from monomer
    vdelta_mA = RpA * vol_A / s.A if s.A > 0 else 0.0
    vdelta_mB = RpB * vol_B / s.B if s.B > 0 else 0.0
    vdelta_PA = RpA * vol_PA / s.nPA if s.nPA > 0 else 0.0
    vdelta_PB = RpB * vol_PB / s.nPB if s.nPB > 0 else 0.0
    vdelta = vdelta_mA + vdelta_mB - vdelta_PA - vdelta_PB
    dydt.V = -s.V * vdelta

    """ Moles of Initiator """
    dydt.I = s.V * (-kd * s.I_conc)

    """ Moles of Free Radical, R """
    dydt.R = s.V * (
        # Formation by initiation
        +2 * f * kd * s.I_conc
        # Consumption by propagation
        - kpRAA * s.R_conc * s.A_conc
        - kpRBB * s.R_conc * s.B_conc
    )

    """ Moles of A Monomer """
    dydt.A = s.V * (
        # Consumption by propagation
        -kpRAA * (s.R_conc + s.RA_conc) * s.A_conc
        - kpRBA * s.RB_conc * s.A_conc
        - kpAA * s.PA_conc * s.A_conc
        - kpBA * s.PB_conc * s.A_conc
        # Formation by depropagation
        + kdAA * s.PAA_conc
        + kdBA * s.PBA_conc
        # Consumption by chain transfer to monomer
        - kf_mAA * s.PA_conc * s.A_conc
        - kf_mBA * s.PB_conc * s.A_conc
        # Consumption by thermal initiation
        - 3 * kthA * s.A_conc**3
    )

    """ Moles of B Monomer """
    dydt.B = s.V * (
        # Consumption by propagation
        -kpRBB * (s.R_conc + s.RB_conc) * s.B_conc
        - kpRAB * s.RA_conc * s.B_conc
        - kpAB * s.PA_conc * s.B_conc
        - kpBB * s.PB_conc * s.B_conc
        # Formation by depropagation
        + kdAB * s.PAB_conc
        + kdBB * s.PBB_conc
        # Consumption by chain transfer to monomer
        - kf_mAB * s.PA_conc * s.B_conc
        - kf_mBB * s.PB_conc * s.B_conc
        # Consumption by thermal initiation
        - 3 * kthB * s.B_conc**3
    )

    """ Moles of R-A dyad"""
    dydt.RA = s.V * (
        # Formation by thermal initiation (3A -> 2RA)
        +2 * kthA * s.A_conc**3
        # Formation by propagation (R + A -> RA)
        + kpRAA * s.R_conc * s.A_conc
        # Consumption by propagation (RA + A -> PAA, RA + B -> PAB)
        - kpRAA * s.RA_conc * s.A_conc
        - kpRAB * s.RA_conc * s.B_conc
        # Formation by chain transfer (PA + A -> PD + RA)
        + kf_mAA * s.PA_conc * s.A_conc
        # Formation by chain transfer (PB + A -> PD + RA)
        + kf_mBA * s.PB_conc * s.A_conc
    )

    """ Total moles of R-B dyad """
    dydt.RB = s.V * (
        # Formation by thermal initiation (3B -> 2RB)
        +2 * kthB * s.B_conc**3
        # Formation by propagation (R + B -> RB)
        + kpRBB * s.R_conc * s.B_conc
        # Consumption by propagation (RB + A -> PBA, RB + B -> PBB)
        - kpRBA * s.RB_conc * s.A_conc
        - kpRBB * s.RB_conc * s.B_conc
        # Formation by chain transfer (PA + B -> PD + RB)
        + kf_mAB * s.PA_conc * s.B_conc
        # Formation by chain transfer (PB + B -> PD + RB)
        + kf_mBB * s.PB_conc * s.B_conc
    )

    c_tAA = c_tBB = 1
    c_tAB = 1

    """ Total moles of ~~AA* polymer """
    dydt.PAA = s.V * (
        # Formation by propagation (RA + A -> PAA)
        +kpRAA * s.RA_conc * s.A_conc
        # Formation by propagation (PBA + A -> PAA)
        + kpAA * s.PBA_conc * s.A_conc
        # Consumption by propagation (PAA + B -> PAB)
        - kpAB * s.PAA_conc * s.B_conc
        # Formation by depropagation (PAAB -> PAA + B)
        + kdAB * s.pAA * s.PAB_conc
        # Consumption by depropagation (PBAA -> PBA + A)
        - kdAA * s.pBA * s.PAA_conc
        # Consumption by chain transfer to monomer (PAA + B -tr-> PD + RB)
        - kf_mAB * s.PAA_conc * s.B_conc
        # Consumption by chain transfer to monomer (PAA + A -tr-> PD + RA)
        - kf_mAA * s.PAA_conc * s.A_conc
        # Consumption by termination (PAA + PA -> (2)PD, PAA + PB -> (2)PD)
        - c_tAA * ktAA * s.PAA_conc * s.PA_conc
        - c_tAB * ktAB * s.PAA_conc * s.PB_conc
    )

    """ Total moles of ~~AB* polymer """
    dydt.PAB = s.V * (
        # Formation by propagation (RA + B -> PAB)
        +kpRAB * s.RA_conc * s.B_conc
        # Formation by propagation (PA + B -> PAB)
        + kpAB * s.PA_conc * s.B_conc
        # Consumption by propagation (PAB + A -> PBA)
        - kpBA * s.PAB_conc * s.A_conc
        # Consumption by propagation (PAB + B -> PBB)
        - kpBB * s.PAB_conc * s.B_conc
        # Formation by depropagation (PABA -> PAB + A)
        + kdBA * s.pAB * s.PBA_conc
        # Formation by depropagation (PABB -> PAB + B)
        + kdBB * s.pAB * s.PBB_conc
        # Consumption by depropagation (PAB -> PA + B)
        - kdAB * s.PAB_conc
        # Consumption by chain transfer to monomer (PAB + A -tr-> PD + RA)
        - kf_mBA * s.PAB_conc * s.A_conc
        # Consumption by chain transfer to monomer (PAB + B -tr-> PD + RB)
        - kf_mBB * s.PAB_conc * s.B_conc
        # Consumption by termination (PAB + PA -> (2)PD, PAB + PB -> (2)PD)
        - c_tAB * ktAB * s.PAB_conc * s.PA_conc
        - c_tBB * ktBB * s.PAB_conc * s.PB_conc
    )

    """ Total moles of ~~BA* polymer """
    dydt.PBA = s.V * (
        # Formation by propagation (RB + A -> PBA)
        +kpRBA * s.RB_conc * s.A_conc
        # Formation by propagation (PB + A -> PBA)
        + kpBA * s.PB_conc * s.A_conc
        # Consumption by propagation (PBA + A -> PAA)
        - kpAA * s.PBA_conc * s.A_conc
        # Consumption by propagation (PBA + B -> PBB)
        - kpAB * s.PBA_conc * s.B_conc
        # Formation by depropagation (PBAA -> PBA + A)
        + kdAA * s.pBA * s.PAA_conc
        # Formation by depropagation (PBAB -> PBA + B)
        + kdAB * s.pBA * s.PAB_conc
        # Consumption by depropagation (PBA -> PB + A)
        - kdBA * s.PBA_conc
        # Consumption by chain transfer to monomer (PBA + A -tr-> PD + RA)
        - kf_mAA * s.PBA_conc * s.A_conc
        # Consumption by chain transfer to monomer (PBA + B -tr-> PD + RB)
        - kf_mAB * s.PBA_conc * s.B_conc
        # Consumption by termination (PBA + PA -> (2)PD, PBA + PB -> (2)PD)
        - c_tAA * ktAA * s.PBA_conc * s.PA_conc
        - c_tAB * ktAB * s.PBA_conc * s.PB_conc
    )

    """ Total moles of ~~BB* polymer """
    dydt.PBB = s.V * (
        # Formation by propagation (RB + B -> PBB)
        +kpRBB * s.RB_conc * s.B_conc
        # Formation by propagation (PAB + B -> PBB)
        + kpBB * s.PAB_conc * s.B_conc
        # Consumption by propagation (PBB + A -> PBA)
        - kpBA * s.PBB_conc * s.A_conc
        # Formation by depropagation (PBBA -> PBB + A)
        + kdBA * s.pBB * s.PBA_conc
        # Consumption by depropagation (PABB -> PAB + B)
        - kdBB * s.pAB * s.PBB_conc
        # Consumption by chain transfer to monomer (PBB + A -tr-> PD + RA)
        - kf_mBA * s.PBB_conc * s.A_conc
        # Consumption by chain transfer to monomer (PBB + B -tr-> PD + RB)
        - kf_mBB * s.PBB_conc * s.B_conc
        # Consumption by termination (PBB + PA -> (2)PD, PBB + PB -> (2)PD)
        - c_tBB * ktBB * s.PBB_conc * s.PB_conc
        - c_tAB * ktAB * s.PBB_conc * s.PA_conc
    )

    """ Total moles of A and B monomer in polymer"""
    dydt.nPA = s.V * RpA
    dydt.nPB = s.V * RpB

    # Formation of polymer from active radicals
    poly_form = (
        kpRAA * s.R_conc * s.A_conc
        + kpRBB * s.R_conc * s.B_conc
        + 2 * kthA * s.A_conc**3
        + 2 * kthB * s.B_conc**3
    )

    """ Active radicals (0th order moment) """
    dydt.lam_0 = s.V * (
        # Formation by propagation
        poly_form
        # Consumption by termination
        - kt * s.lam_0_conc**2
    )

    """ Active radicals (1st order moment) """
    dydt.lam_1 = s.V * (
        # Formation by propagation
        poly_form
        # Formation by propagation
        + kp * s.M_conc * s.lam_0_conc
        # Consumption by termination
        - kt * s.lam_0_conc * s.lam_1_conc
        # Consumption from chain transfer
        - kfm * s.M_conc * (s.lam_1_conc - s.lam_0_conc)
    )

    """Active radicals (2nd order moment)"""
    dydt.lam_2 = s.V * (
        # Formation by propagation
        poly_form
        # Formation by propagation
        + kp * s.M_conc * (s.lam_0_conc + 2 * s.lam_1_conc)
        # Consumption by termination
        - kt * s.lam_0_conc * s.lam_2_conc
        # Consumption from chain transfer
        - kfm * s.M_conc * (s.lam_2_conc - s.lam_0_conc)
    )

    """Dead radicals (0th order moment)"""
    dydt.mu_0 = s.V * (
        # Formation by termination
        +(0.5 * ktc + ktd) * s.lam_0_conc**2
        # Formation by chain transfer
        + kfm * s.M_conc * s.lam_0_conc
    )

    """Dead radicals (1st order moment)"""
    dydt.mu_1 = s.V * (
        # Formation by termination
        +(ktc + ktd) * s.lam_0_conc * s.lam_1_conc
        # Formation by chain transfer
        + kfm * s.M_conc * s.lam_1_conc
    )

    """Dead radicals (2nd order moment)"""
    dydt.mu_2 = s.V * (
        # Formation by termination (comb.)
        +ktc * (s.lam_0_conc * s.lam_2_conc + s.lam_1_conc**2)
        # Formation by termination (disp.)
        + ktd * s.lam_0_conc * s.lam_2_conc
        # Formation by chain transfer
        + kfm * s.M_conc * s.lam_2_conc
    )

    return dydt


def sequence_odes(
    dydt: SequenceModelState,
    k: RateCoefficients,
    t: NDArrayType,
    sms: SequenceModelState,
    s: ModelState,
) -> SequenceModelState:

    dydt.t = t

    kpAA, kpAB, kpBA, kpBB = k.kp()
    kdAA, kdAB, kdBA, kdBB = k.kd()
    ktAA, ktAB, ktBB = get_kt_from_Cheong_2003(s, k)

    kf_mAA, kf_mAB, kf_mBA, kf_mBB = k.kfm()

    aSA0_conc = sms.aSA0 / s.V
    aSB0_conc = sms.aSB0 / s.V
    aSA1_conc = sms.aSA1 / s.V
    aSB1_conc = sms.aSB1 / s.V
    aSA2_conc = sms.aSA2 / s.V
    aSB2_conc = sms.aSB2 / s.V

    # Rates of termination and chain transfer to monomer
    # Ctrij_x = c_tij * ktij * sms.aSijx * sms.aSij0 + c_tij * kf_mij * sms.aSijx * s.j
    RtAA_0 = ktAA * aSA0_conc * aSA0_conc + kf_mAA * aSA0_conc * s.A_conc
    RtAB_0 = ktAB * aSA0_conc * aSB0_conc + kf_mAB * aSA0_conc * s.B_conc
    RtBA_0 = ktAB * aSB0_conc * aSA0_conc + kf_mBA * aSB0_conc * s.A_conc
    RtBB_0 = ktBB * aSB0_conc * aSB0_conc + kf_mBB * aSB0_conc * s.B_conc
    RtAA_1 = ktAA * aSA1_conc * aSA0_conc + kf_mAA * aSA1_conc * s.A_conc
    RtAB_1 = ktAB * aSA1_conc * aSB0_conc + kf_mAB * aSA1_conc * s.B_conc
    RtBA_1 = ktAB * aSB1_conc * aSA0_conc + kf_mBA * aSB1_conc * s.A_conc
    RtBB_1 = ktBB * aSB1_conc * aSB0_conc + kf_mBB * aSB1_conc * s.B_conc
    RtAA_2 = ktAA * aSA2_conc * aSA0_conc + kf_mAA * aSA2_conc * s.A_conc
    RtAB_2 = ktAB * aSA2_conc * aSB0_conc + kf_mAB * aSA2_conc * s.B_conc
    RtBA_2 = ktAB * aSB2_conc * aSA0_conc + kf_mBA * aSB2_conc * s.A_conc
    RtBB_2 = ktBB * aSB2_conc * aSB0_conc + kf_mBB * aSB2_conc * s.B_conc

    a10_frac = aSA1_conc / aSA0_conc if aSA0_conc > 0 else 0
    b10_frac = aSB1_conc / aSB0_conc if aSB0_conc > 0 else 0
    a20_frac = aSA2_conc / aSA0_conc if aSA0_conc > 0 else 0
    b20_frac = aSB2_conc / aSB0_conc if aSB0_conc > 0 else 0
    kthA = k.mpropsA.kth
    kthB = k.mpropsB.kth

    # Active A sequences (0th order moment)
    dydt.aSA0 = s.V * (
        # Formation by propagation (PA + A -> PA, PB + A -> PA)
        +kpAA * s.A_conc * (s.R_conc + aSA0_conc)
        + kpBA * s.A_conc * (s.RB_conc + aSB0_conc)
        + kthA * s.A_conc**3
        # Consumption by propagation (PA + A -> PA, PA + B -> PB)
        - kpAA * s.A_conc * aSA0_conc
        - kpAB * s.B_conc * (s.RA_conc + aSA0_conc)
        # Formation by depropagation (PAA -> PA + A, PAB -> PA + B)
        + kdAA * s.pAA * aSA0_conc
        + kdAB * s.pAB * aSB0_conc
        # Consumption by depropagation (PAA -> PA + A, PBA -> PB + A)
        - kdAA * s.pAA * aSA0_conc
        - kdBA * s.pBA * aSA0_conc
        # Consumption by termination (PA + PA -> D, PA + PB -> D)
        - RtAA_0
        - RtAB_0
    )

    # Active B sequences (0th order moment)
    dydt.aSB0 = s.V * (
        # Formation by propagation (PB + B -> PB, PA + B -> PB)
        +kpBB * s.B_conc * (s.R_conc + aSB0_conc)
        + kpAB * s.B_conc * (s.RA_conc + aSA0_conc)
        + kthB * s.B_conc**3
        # Consumption by propagation (PB + B -> PB, PB + A -> PA)
        - kpBB * s.B_conc * aSB0_conc
        - kpBA * s.A_conc * (s.RB_conc + aSB0_conc)
        # Formation by depropagation (PBB -> PB + B, PBA -> PB + A)
        + kdBB * s.pBB * aSB0_conc
        + kdBA * s.pBA * aSA0_conc
        # Consumption by depropagation (PBB -> PB + B, PAB -> PA + B)
        - kdBB * s.pBB * aSB0_conc
        - kdAB * s.pAB * aSB0_conc
        # Consumption by termination (PB + PB -> D, PB + PA -> D)
        - RtBB_0
        - RtBA_0
    )

    # Active A sequences (1st order moment)
    dydt.aSA1 = s.V * (
        # Formation by propagation (PA + A -> PA, PB + A -> PA)
        +kpAA * s.A_conc * (s.R_conc + aSA0_conc + aSA1_conc)
        + kpBA * s.A_conc * (s.RB_conc + aSB0_conc)
        # Consumption by propagation (PA + A -> PA, PA + B -> PB)
        - kpAA * s.A_conc * aSA1_conc
        - kpAB * s.B_conc * (s.RA_conc + aSA1_conc)
        # Formation by depropagation (PAA -> PA + A, PAB -> PA + B)
        + kdAA * s.pAA * (aSA1_conc - aSA0_conc)
        + kdAB * s.pAB * a10_frac * aSB0_conc
        # Consumption by depropagation (PAA -> PA + A, PBA -> PB + A)
        - kdAA * s.pAA * aSA1_conc
        - kdBA * s.pBA * aSA0_conc
        # Consumption by termination (PA + PA -> D, PA + PB -> D)
        - RtAA_1
        - RtAB_1
    )

    # Active B sequences (1st order moment)
    dydt.aSB1 = s.V * (
        # Formation by propagation (PB + B -> PB, PA + B -> PB)
        +kpBB * s.B_conc * (s.R_conc + aSB0_conc + aSB1_conc)
        + kpAB * s.B_conc * (s.RA_conc + aSA0_conc)
        # Consumption by propagation (PB + B -> PB, PB + A -> PA)
        - kpBB * s.B_conc * aSB1_conc
        - kpBA * s.A_conc * (s.RB_conc + aSB1_conc)
        # Formation by depropagation (PBB -> PB + B, PBA -> PB + A)
        + kdBB * s.pBB * (aSB1_conc - aSB0_conc)
        + kdBA * s.pBA * b10_frac * aSA0_conc
        # Consumption by depropagation (PBB -> PB + B, PAB -> PA + B)
        - kdBB * s.pBB * aSB1_conc
        - kdAB * s.pAB * aSB0_conc
        # Consumption by termination (PB + PB -> D, PB + PA -> D)
        - RtBB_1
        - RtBA_1
    )

    # return dydt

    # Active A sequences (2nd order moment)
    dydt.aSA2 = s.V * (
        # Formation by propagation (PA + A -> PA, PB + A -> PA)
        +kpAA * s.A_conc * (s.R_conc + aSA0_conc + 2 * aSA1_conc + aSA2_conc)
        + kpBA * s.A_conc * (s.RB_conc + aSB0_conc)
        # Consumption by propagation (PA + A -> PA, PA + B -> PB)
        - kpAA * s.A_conc * aSA2_conc
        - kpAB * s.B_conc * (s.RA_conc + aSA2_conc)
        # Formation by depropagation (PAA -> PA + A, PAB -> PA + B)
        + kdAA * s.pAA * (aSA2_conc - 2 * aSA1_conc + aSA0_conc)
        + kdAB * s.pAB * a20_frac * aSB0_conc
        # Consumption by depropagation (PAA -> PA + A, PBA -> PB + A)
        - kdAA * s.pAA * aSA2_conc
        - kdBA * s.pBA * aSA0_conc
        # Consumption by termination (PA + PA -> D, PA + PB -> D)
        - RtAA_2
        - RtAB_2
    )

    # Active B sequences (2nd order moment)
    dydt.aSB2 = s.V * (
        # Formation by propagation (PB + B -> PB, PA + B -> PB)
        +kpBB * s.B_conc * (s.R_conc + aSB0_conc + 2 * aSB1_conc + aSB2_conc)
        + kpAB * s.B_conc * (s.RA_conc + aSA0_conc)
        # Consumption by propagation (PB + B -> PB, PB + A -> PA)
        - kpBB * s.B_conc * aSB2_conc
        - kpBA * s.A_conc * (s.RB_conc + aSB2_conc)
        # Formation by depropagation (PBB -> PB + B, PBA -> PB + A)
        + kdBB * s.pBB * (aSB2_conc - 2 * aSB1_conc + aSB0_conc)
        + kdBA * s.pBA * b20_frac * aSA0_conc
        # Consumption by depropagation (PBB -> PB + B, PAB -> PA + B)
        - kdBB * s.pBB * aSB2_conc
        - kdAB * s.pAB * aSB0_conc
        # Consumption by termination (PB + PB -> D, PB + PA -> D)
        - RtBB_2
        - RtBA_2
    )

    # Inactive A sequences (0th order moment)
    dydt.iSA0 = s.V * (
        # Formation by cross-propagation (PA + B -> PB)
        +kpAB * s.B_conc * (s.RA_conc + aSA0_conc)
        # Consumption by depropagation (PAB -> PA + B)
        - kdAB * s.pAB * aSB0_conc
        # Formation by termination (PA + PA -> D, PA + PB -> D)
        + RtAA_0
        + RtAB_0
    )

    # Inactive B sequences (0th order moment)
    dydt.iSB0 = s.V * (
        # Formation by cross-propagation (PB + A -> PA)
        +kpBA * s.A_conc * (s.RB_conc + aSB0_conc)
        # Consumption by depropagation (PBA -> PB + A)
        - kdBA * s.pBA * aSA0_conc
        # Formation by termination (PB + PB -> D, PB + PA -> D)
        + RtBB_0
        + RtBA_0
    )

    # Inactive A sequences (1st order moment)
    dydt.iSA1 = s.V * (
        # Formation by cross-propagation (PA + B -> PB)
        +kpAB * s.B_conc * (s.RA_conc + aSA1_conc)
        # Consumption by depropagation (PAB -> PA + B)
        - kdAB * s.pAB * a10_frac * aSB0_conc
        # Formation by termination (PA + PA -> D, PA + PB -> D)
        + RtAA_1
        + RtAB_1
    )

    # Inactive B sequences (1st order moment)
    dydt.iSB1 = s.V * (
        # Formation by cross-propagation (PB + A -> PA)
        +kpBA * s.A_conc * (s.RB_conc + aSB1_conc)
        # Consumption by depropagation (PBA -> PB + A)
        - kdBA * s.pBA * b10_frac * aSA0_conc
        # Formation by termination (PB + PB -> D, PB + PA -> D)
        + RtBB_1
        + RtBA_1
    )

    # Inactive A sequences (2nd order moment)
    dydt.iSA2 = s.V * (
        # Formation by cross-propagation (PA + B -> PB)
        +kpAB * s.B_conc * (s.RA_conc + aSA2_conc)
        # Consumption by depropagation (PAB -> PA + B)
        - kdAB * s.pAB * a20_frac * aSB0_conc
        # Formation by termination (PA + PA -> D, PA + PB -> D)
        + RtAA_2
        + RtAB_2
    )

    # Inactive B sequences (2nd order moment)
    dydt.iSB2 = s.V * (
        # Formation by cross-propagation (PB + A -> PA)
        +kpBA * s.A_conc * (s.RB_conc + aSB2_conc)
        # Consumption by depropagation (PBA -> PB + A)
        - kdBA * s.pBA * b20_frac * aSA0_conc
        # Formation by termination (PB + PB -> D, PB + PA -> D)
        + RtBB_2
        + RtBA_2
    )

    return dydt



from typing import Callable, Dict, Type, TypeAlias, Generic, TypeVar
from enum import Enum
from dataclasses import dataclass

import numpy as np

from sparks.core import NDArrayType
from .state import StateType

# Type variables for generic types
RateType = TypeVar("RateType")
RxnIDType = TypeVar("RxnIDType", bound=Enum)


@dataclass
class Reaction(Generic[RateType, StateType]):
    """
    Represents a single chemical reaction with reactants, products, and a rate law.

    Args:
        reactants: Dict mapping species names to their stoichiometric coefficients.
        products: Dict mapping species names to their stoichiometric coefficients.
        rate: Function that calculates the rate of the reaction given the rate coefficients and state.
            def rate_fn(k: RateType, state: StateType) -> NDArrayType:
                pass
    """

    reactants: Dict[str, float]
    products: Dict[str, float]
    rate: Callable[[RateType, StateType], NDArrayType]


def make_reaction(rate_cls: Type[RateType], state_cls: Type[StateType]):
    """Factory function that returns a properly typed Reaction constructor"""

    def typed_reaction(
        reactants: Dict[str, float],
        products: Dict[str, float],
        rate: Callable[[RateType, StateType], NDArrayType],
    ) -> Reaction[RateType, StateType]:
        return Reaction(reactants=reactants, products=products, rate=rate)

    return typed_reaction


# Type alias for a reaction network definition
ReactionNetworkDef: TypeAlias = Dict[RxnIDType, Reaction[RateType, StateType]]


def calculate_rates(
    state: StateType,
    k: RateType,
    reactions: ReactionNetworkDef,
) -> Dict[RxnIDType, NDArrayType]:
    """
    Calculate the rate of each reaction in the system.

    Args:
        state: Current system state.
        k: Rate coefficients.
        reactions: Reaction network definition.

    Returns:
        Dict mapping reaction IDs to their rates.
    """
    return {rxn_id: rxn.rate(k, state) for rxn_id, rxn in reactions.items()}


def calculate_probabilities(
    state: StateType,
    k: RateType,
    reactions: ReactionNetworkDef,
) -> Dict[RxnIDType, NDArrayType]:
    """
    Calculate the probability of each reaction occurring.

    Probabilities are normalized such that they sum to 100%.

    Args:
        state: Current system state.
        k: Rate coefficients
        reactions: Reaction network definition.

    Returns:
        Dict mapping reaction IDs to their probabilities (0-100%).
    """
    rates = calculate_rates(state, k, reactions)

    total_rate = np.zeros_like(next(iter(rates.values())), dtype=float)
    for rate in rates.values():
        total_rate += rate

    # Convert to probabilities
    return {
        rxn_id: np.where(total_rate > 0, 100 * rate / total_rate, 0.0)
        for rxn_id, rate in rates.items()
    }


def validate_reaction_network(
    state_class: Type[StateType],
    reactions: ReactionNetworkDef,
) -> None:
    """
    Validate that all species in reactions exist in the state class.

    Args:
        state_class: State class with species as fields.
        reactions: Reaction network to validate.

    Raises:
        ValueError: If any species in reactions don't exist in the state class.
    """

    valid_species = set(state_class.get_species_names())

    for rxn_id, reaction in reactions.items():

        # Check reactants
        invalid_reactants = set(reaction.reactants.keys()) - valid_species
        if invalid_reactants:
            raise ValueError(
                f"Reaction {rxn_id} has invalid reactants: {invalid_reactants}. "
                f"Valid species are: {valid_species}."
            )

        # Check products
        invalid_products = set(reaction.products.keys()) - valid_species
        if invalid_products:
            raise ValueError(
                f"Reaction {rxn_id} has invalid products: {invalid_products}. "
                f"Valid species are: {valid_species}."
            )



from .base import Simulator

from .deterministic import (
    DeterministicSimulator,
    default_solver_params,
)

from .stochastic import StochasticSimulator

__all__ = [
    "Simulator",
    "DeterministicSimulator",
    "default_solver_params",
    "StochasticSimulator",
]



from runkmc.kmc import RunKMC
from runkmc.results import SimulationResult
from .base import KMCConfig, StochasticSimulator


__all__ = ["KMCConfig", "RunKMC", "SimulationResult", "StochasticSimulator"]



from typing import Type
from pathlib import Path

import numpy as np
from runkmc.kmc.config import KMCConfig
from runkmc.kmc import RunKMC, SimulationConfig
from runkmc.results import SimulationResult

from sparks.models import (
    RateType,
    StateType,
    SimulationData,
    SupportsStochasticSimulation,
)
from sparks.simulators import Simulator


class StochasticSimulator(Simulator):
    """Kinetic Monte Carlo Simulator (using RunKMC)"""

    def __init__(
        self,
        model: SupportsStochasticSimulation,
        k: RateType,
        data_dir: Path | str,
        compile: bool = False,
    ):

        super().__init__(model, k)
        self.model = model

        self.model_state_class: Type[StateType] = self.model.StateClass
        self.run_kmc = RunKMC(data_dir, compile=compile)

    def simulate(
        self,
        t_eval: np.ndarray,
        init_state: StateType,
        num_units: int,
        model_name: str,
        **kwargs,
    ) -> SimulationData[StateType]:

        self.y0 = init_state
        self.t_eval = t_eval
        self.model_name = model_name

        self.kmc_config = KMCConfig(
            num_units=num_units,
            termination_time=t_eval[-1],
            analysis_time=t_eval[1] - t_eval[0],
        )
        kmc_inputs = self.model.get_kmc_inputs(
            self.k, self.y0, self.kmc_config, model_name=model_name, **kwargs
        )

        self.sim_config = SimulationConfig(model_name, kmc_inputs, **kwargs)

        self.result: SimulationResult = self.run_kmc.run_from_config(self.sim_config)

        data, mdata = self.model.parse_kmc_outputs(self.result, self.model_name)
        self.mdata = mdata

        return data
